package eu.jonahbauer.tichu.common.model;

import eu.jonahbauer.tichu.common.model.combinations.*;
import eu.jonahbauer.tichu.common.model.exceptions.IncompatibleCombinationException;
import eu.jonahbauer.tichu.common.model.exceptions.InvalidCombinationException;
import eu.jonahbauer.tichu.common.model.exceptions.TooLowCombinationException;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static eu.jonahbauer.tichu.common.model.Card.*;
import static eu.jonahbauer.tichu.common.model.CombinationType.*;
import static java.util.function.Predicate.not;
import static org.junit.jupiter.api.Assertions.*;

public class CombinationTest {
    private static final Map<CombinationType, List<List<List<Card>>>> VALID_COMBINATIONS = Map.of(
            SINGLE, Stream.concat(
                    Arrays.stream(Card.values()) // [[RED_2]], [[GREEN_2]], [[BLUE_2]], ...
                            .map(List::of)
                            .map(List::of),
                    Arrays.stream(Suit.values()) // [[RED_2], [RED_3], [RED_4], ...], [[BLUE_2], ...], ...
                            .map(suit -> IntStream.rangeClosed(2, 14)
                                    .mapToObj(Card::getCardsByValue)
                                    .flatMap(cards -> cards.stream()
                                            .filter(card -> card.suit() == suit)
                                    )
                                    .map(List::of)
                                    .toList()
                            )
            ).toList(),
            PAIR, List.of(List.of(
                    List.of(BLUE_5, RED_5),
                    List.of(BLUE_JACK, BLACK_JACK),
                    List.of(PHOENIX, GREEN_KING),
                    List.of(RED_ACE, PHOENIX)
            )),
            TRIPLE, List.of(List.of(
                    List.of(BLUE_5, RED_5, GREEN_5),
                    List.of(BLUE_7, RED_7, PHOENIX),
                    List.of(BLUE_10, PHOENIX, BLACK_10),
                    List.of(PHOENIX, BLUE_QUEEN, RED_QUEEN)
            )),
            FULL_HOUSE, List.of(List.of(
                    List.of(BLUE_5, RED_5, GREEN_5, BLACK_10, GREEN_10),
                    List.of(BLUE_6, RED_6, GREEN_6, BLACK_4, PHOENIX),
                    List.of(BLUE_8, RED_8, GREEN_8, PHOENIX, GREEN_4),
                    List.of(BLUE_9, RED_9, PHOENIX, BLACK_ACE, GREEN_ACE),
                    List.of(BLUE_JACK, PHOENIX, GREEN_JACK, BLACK_4, GREEN_4),
                    List.of(PHOENIX, RED_KING, GREEN_KING, BLACK_4, GREEN_4)
            )),
            SEQUENCE, List.of(
                    List.of(
                            List.of(MAHJONG, RED_2, BLACK_3, RED_4, BLUE_5)
                    ),
                    List.of(
                            List.of(MAHJONG, PHOENIX, BLACK_3, RED_4, BLUE_5)
                    ),
                    List.of(
                            List.of(MAHJONG, RED_2, BLACK_3, PHOENIX, BLUE_5),
                            List.of(BLUE_4, RED_5, BLACK_6, RED_7, BLUE_8),
                            List.of(PHOENIX, RED_6, BLACK_7, RED_8, BLUE_9),
                            List.of(BLUE_10, RED_JACK, BLACK_QUEEN, RED_KING, PHOENIX)
                    ),
                    List.of(
                            List.of(RED_6, BLUE_7, BLUE_8, RED_9, BLACK_10, RED_JACK, BLUE_QUEEN),
                            List.of(PHOENIX, BLUE_8, BLUE_9, RED_10, BLACK_JACK, RED_QUEEN, BLUE_KING),
                            List.of(RED_8, BLUE_9, BLUE_10, RED_JACK, BLACK_QUEEN, RED_KING, PHOENIX)
                    )
            ),
            PAIR_SEQUENCE, List.of(
                    List.of(
                            List.of(RED_2, BLUE_2, GREEN_3, BLACK_3),
                            List.of(RED_3, BLUE_3, PHOENIX, BLACK_4)
                    ),
                    List.of(
                            List.of(RED_2, BLUE_2, GREEN_3, BLACK_3, BLACK_4, RED_4),
                            List.of(RED_5, BLUE_5, PHOENIX, BLACK_6, BLACK_7, RED_7)
                    )
            ),
            BOMB, List.of(
                    IntStream.rangeClosed(2, 14).mapToObj(Card::getCardsByValue).map(List::copyOf).toList()
            ),
            BOMB_SEQUENCE, List.of(List.of(
                    List.of(RED_2, RED_3, RED_4, RED_5, RED_6),
                    List.of(RED_2, RED_3, RED_4, RED_5, RED_6, RED_7, RED_8, RED_9),
                    List.of(BLUE_7, BLUE_8, BLUE_9, BLUE_10, BLUE_JACK, BLUE_QUEEN, BLUE_KING, BLUE_ACE)
            ))
    );

    private static final Map<CombinationType, List<List<Card>>> INVALID_COMBINATIONS = Map.of(
            SINGLE, List.of(
                    List.of(BLUE_2, RED_2), // too many cards
                    List.of(BLUE_2, RED_3) // too many cards
            ),
            PAIR, List.of(
                    List.of(RED_ACE, BLUE_ACE, GREEN_ACE), // too many cards
                    List.of(RED_ACE, RED_ACE), // duplicate card
                    List.of(RED_KING, RED_ACE), // different cards
                    List.of(MAHJONG, PHOENIX), // phoenix cannot be mahjong / mahjong cannot be part of a tuple
                    List.of(DRAGON, PHOENIX), // phoenix cannot be dragon / dragon cannot be part of a tuple
                    List.of(HOUND, PHOENIX), // phoenix cannot be hound / hound cannot be part of a tuple
                    List.of(RED_5) // too few cards
            ),
            TRIPLE, List.of(
                    List.of(RED_ACE, BLUE_ACE, GREEN_ACE, BLACK_ACE), // too many cards
                    List.of(RED_ACE, BLUE_ACE, BLUE_ACE), // duplicate card
                    List.of(RED_ACE, BLUE_ACE, GREEN_KING), // different cards
                    List.of(MAHJONG, DRAGON, PHOENIX),
                    List.of(RED_ACE, BLUE_ACE), // too few cards
                    List.of(RED_ACE) // too few cards
            ),
            FULL_HOUSE, List.of(
                    List.of(BLUE_5, RED_5, BLUE_5, BLACK_4, GREEN_4, BLUE_4), // too many cards
                    List.of(BLUE_5, RED_5, BLUE_5, BLACK_4), // too few cards
                    List.of(BLUE_5, RED_5, BLUE_4, BLACK_4, GREEN_4), // wrong order
                    List.of(BLUE_5, RED_5, BLUE_6, BLACK_4, GREEN_4), // different cards in triple
                    List.of(BLUE_5, RED_5, GREEN_5, BLACK_3, GREEN_4), // different cards in pair
                    List.of(BLUE_5, RED_5, GREEN_5, BLUE_5, GREEN_5), // duplicate cards
                    List.of(BLUE_5, PHOENIX, GREEN_5, PHOENIX, RED_5), // duplicate cards
                    List.of(BLUE_5, RED_5, GREEN_5, DRAGON, GREEN_4), // dragon cannot be part of a tuple
                    List.of(BLUE_5, RED_5, GREEN_5, HOUND, GREEN_4), // hound cannot be part of a tuple
                    List.of(BLUE_5, RED_5, GREEN_5, MAHJONG, PHOENIX) // mahjong cannot be part of a tuple
            ),
            SEQUENCE, List.of(
                    List.of(RED_ACE, RED_2, BLACK_3, RED_4, BLUE_5), // ace is no 1
                    List.of(RED_2, BLACK_3, RED_4, BLUE_5), // too few cards
                    List.of(RED_2, BLACK_3, RED_4), // too few cards
                    List.of(RED_2, BLACK_3), // too few cards
                    List.of(RED_2), // too few cards
                    List.of(PHOENIX, MAHJONG, RED_2, BLACK_3, PHOENIX, BLUE_5), // phoenix cannot be 0
                    List.of(RED_2, MAHJONG, BLACK_4, RED_5, GREEN_6), // wrong order
                    List.of(BLACK_8, GREEN_7, BLUE_6, RED_5, GREEN_4), // wrong order
                    List.of(BLACK_10, GREEN_JACK, BLUE_QUEEN, RED_KING, GREEN_ACE, DRAGON), // dragon cannot be part of sequence
                    List.of(BLACK_10, GREEN_JACK, BLUE_QUEEN, RED_KING, GREEN_ACE, PHOENIX) // phoenix cannot be higher than ace
            ),
            PAIR_SEQUENCE, List.of(
                    List.of(RED_2, BLUE_2), // too few cards
                    List.of(RED_2, BLUE_2, GREEN_3), // too few cards
                    List.of(RED_2, BLUE_3, GREEN_4, BLUE_5), // missing pairs
                    List.of(RED_2, BLUE_2, GREEN_3, BLUE_3, BLACK_4), // partially missing pairs
                    List.of(RED_3, BLUE_3, GREEN_2, BLUE_2), // wrong order
                    List.of(PHOENIX, MAHJONG, RED_2, BLUE_2), // phoenix cannot be mahjong
                    List.of(RED_ACE, BLUE_ACE, DRAGON, PHOENIX) // phoenix cannot be dragon
            ),
            BOMB, List.of(
                    List.of(RED_2, RED_2, BLUE_2), // too few cards
                    List.of(RED_2, RED_2, BLUE_2, GREEN_2, BLACK_2), // too many cards / duplicate cards
                    List.of(RED_2, BLUE_2, GREEN_2, RED_2), // duplicate cards
                    List.of(RED_2, BLUE_2, GREEN_2, RED_3), // different cards
                    List.of(RED_2, BLUE_2, GREEN_2, PHOENIX) // phoenix cannot be part of a bomb
            ),
            BOMB_SEQUENCE, List.of(
                    List.of(RED_2, RED_3, RED_4, BLACK_5, RED_6), // wrong suit
                    List.of(BLUE_10, PHOENIX, BLUE_QUEEN, BLUE_KING, BLUE_ACE), // phoenix cannot be part of a bomb
                    List.of(MAHJONG, RED_2, RED_3, RED_4, RED_5), // mahjong cannot be part of a bomb
                    List.of(RED_2, RED_3, RED_4, RED_5), // too few cards
                    List.of(RED_2, RED_3, RED_4, RED_5, PHOENIX), // phoenix cannot be part of a bomb
                    List.of(RED_2, RED_3, RED_5, RED_4, RED_6), // wrong order
                    List.of(RED_6, RED_5, RED_4, RED_3, RED_2) // wrong order
            )
    );

    @ParameterizedTest
    @EnumSource(CombinationType.class)
    public void testCheck(CombinationType type) {
        VALID_COMBINATIONS.get(type).stream().flatMap(List::stream).forEach(cards -> assertValidCombination(type, cards));
        INVALID_COMBINATIONS.get(type).forEach(cards -> assertInvalidCombination(type, cards));
    }

    @ParameterizedTest
    @EnumSource(CombinationType.class)
    public void testIsCompatible(CombinationType type) {
        VALID_COMBINATIONS.get(type).forEach(combinations -> {
            for (List<Card> combinationA : combinations) {
                for (List<Card> combinationB : combinations) {
                    assertIsCompatible(type, combinationA,  combinationB);
                }
            }
        });
    }

    @ParameterizedTest
    @EnumSource(CombinationType.class)
    public void testIsHigher(CombinationType type) {
        VALID_COMBINATIONS.get(type).forEach(combinations -> {
            Stack stack = Stack.empty();
            for (int i = 0; i < combinations.size(); i++) {
                stack = stack.put(Combination.of(type, combinations.get(i)));
                for (int j = 0; j < i; j++) {
                    assertIsNotHigher(stack, Combination.of(type, combinations.get(j)));
                }

                for (int j = i + 1; j < combinations.size(); j++) {
                    assertIsHigher(stack, Combination.of(type, combinations.get(j)));
                }
            }
        });
    }

    @Test
    public void testBombIsCompatible() {
        List<Stack> stacks = VALID_COMBINATIONS.entrySet().stream()
                .flatMap(entry -> entry.getValue().stream()
                        .flatMap(List::stream)
                        .filter(not(cards -> cards.size() == 1 && cards.get(0) == HOUND)) // hound cannot be bombed
                        .map(cards -> Combination.of(entry.getKey(), cards))
                )
                .map(Stack::of)
                .toList();
        Stack hound = Stack.of(Single.of(HOUND));

        VALID_COMBINATIONS.get(BOMB).stream()
                .flatMap(List::stream)
                .map(BombTuple::of)
                .forEach(bomb -> {
                    stacks.forEach(stack -> assertIsCompatible(stack, bomb));
                    assertIsIncompatible(hound, bomb);
                });

        VALID_COMBINATIONS.get(BOMB_SEQUENCE).stream()
                .flatMap(List::stream)
                .map(BombSequence::of)
                .forEach(bomb -> {
                    stacks.forEach(stack -> assertIsCompatible(stack, bomb));
                    assertIsIncompatible(hound, bomb);
                });
    }

    private static void assertValidCombination(CombinationType type, List<Card> cards) {
        try {
            Combination.of(type, cards);
        } catch (InvalidCombinationException e) {
            fail("%s should be a valid %s.".formatted(cards, type), e);
        }

        assertTrue(
                type.check(cards),
                "%s should be a valid %s.".formatted(cards, type)
        );
    }

    private static void assertInvalidCombination(CombinationType type, List<Card> cards) {
        assertThrows(InvalidCombinationException.class, () -> {
            Combination.of(type, cards);
        }, "%s should not be a valid %s.".formatted(cards, type));

        assertFalse(
                type.check(cards),
                "%s should not be a valid %s.".formatted(cards, type)
        );
    }

    private static void assertIsCompatible(CombinationType type, List<Card> stack, List<Card> cards) {
        assertIsCompatible(Stack.of(Combination.of(type, stack)), Combination.of(type, cards));
    }

    private static void assertIsCompatible(Stack stack, Combination combination) {
        assertTrue(
                combination.isCompatibleWith(stack),
                "%s should be compatible with %s".formatted(combination, stack)
        );

        assertTrue(
                stack.isCompatible(combination),
                "%s should be compatible with %s".formatted(combination, stack)
        );
    }

    private static void assertIsIncompatible(Stack stack, Combination combination) {
        assertFalse(
                stack.isCompatible(combination),
                "%s should be incompatible with %s".formatted(combination, stack)
        );

        assertFalse(
                stack.isCompatibleAndHigher(combination),
                "%s should be incompatible with %s".formatted(combination, stack)
        );

        assertFalse(
                combination.isCompatibleWith(stack),
                "%s should be incompatible with %s".formatted(combination, stack)
        );

        assertFalse(
                combination.isCompatibleAndHigher(stack),
                "%s should be incompatible with %s".formatted(combination, stack)
        );

        assertThrows(IncompatibleCombinationException.class, () -> {
            //noinspection ResultOfMethodCallIgnored
            stack.put(combination);
        }, "%s should be incompatible with %s".formatted(combination, stack));
    }

    private static void assertIsHigher(Stack stack, Combination combination) {
        assertTrue(
                combination.isHigherThan(stack),
                "%s should be higher than %s".formatted(combination, stack)
        );

        assertTrue(
                stack.isHigher(combination),
                "%s should be higher than %s".formatted(combination, stack)
        );

        assertTrue(
                combination.isCompatibleAndHigher(stack),
                "%s should be higher than %s".formatted(combination, stack)
        );

        assertTrue(
                stack.isCompatibleAndHigher(combination),
                "%s should be higher than %s".formatted(combination, stack)
        );
    }

    private static void assertIsNotHigher(Stack stack, Combination combination) {
        assertFalse(
                combination.isHigherThan(stack),
                "%s should not be higher than %s".formatted(combination, stack)
        );

        assertFalse(
                stack.isHigher(combination),
                "%s should not be higher than %s".formatted(combination, stack)
        );

        assertFalse(
                combination.isCompatibleAndHigher(stack),
                "%s should not be higher than %s".formatted(combination, stack)
        );

        assertFalse(
                stack.isCompatibleAndHigher(combination),
                "%s should not be higher than %s".formatted(combination, stack)
        );

        assertThrows(TooLowCombinationException.class, () -> {
            //noinspection ResultOfMethodCallIgnored
            stack.put(combination);
        });
    }
}
